import argparse
import re
import sys
from typing import TYPE_CHECKING

import git

from . import GitMeTheURL, GMTUException

if TYPE_CHECKING:
    from typing import Optional, Tuple, Union

def main() -> None:
    #----------------------------------
    # Collect command line arguments
    #----------------------------------
    parser = argparse.ArgumentParser(
        description="Create a shareable URL to a Git repository"
    )
    parser.add_argument(
        '-e', '--exact', dest='exact_commit', action='store_true',default=False,
        help="If set, URL will point to the file at the exact commit."
    )
    parser.add_argument('target_path',
        nargs="?", default=".",
        help='Generate a public URL to this path.'
    )
    options = parser.parse_args()

    #----------------------------------
    # Parse target path
    m = re.fullmatch(r'(?P<path>.+?):(?P<line>\d+)(?:-(?P<to_line>\d+))?', options.target_path)
    line = None # type: Optional[Union[int, Tuple[int, int]]]
    if not m:
        path = options.target_path
    else:
        path = m.group("path")
        line = int(m.group("line"))
        to_line_str = m.group("to_line")
        if to_line_str is not None:
            to_line = int(to_line_str)
            line = (
                min(line, to_line),
                max(line, to_line),
            )

    #----------------------------------
    gmtu = GitMeTheURL()

    #----------------------------------
    # Translate the path!
    #----------------------------------
    try:
        url = gmtu.get_source_url(path, line=line, exact_commit=options.exact_commit)
    except git.InvalidGitRepositoryError:
        print("error: Path does not point to a git repository: %s" % path, file=sys.stderr)
        sys.exit(1)
    except git.NoSuchPathError:
        print("error: Path does not exist: %s" % path, file=sys.stderr)
        sys.exit(1)
    except GMTUException as e:
        print("error:", e.args[0], file=sys.stderr)
        sys.exit(1)

    #----------------------------------
    # Output
    #----------------------------------
    print(url)



import sys
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from importlib.metadata import EntryPoint

if sys.version_info >= (3,10,0):
    from importlib import metadata

    def _get_entry_points(group_name: str) -> List['EntryPoint']:
        eps = []
        for ep in metadata.entry_points().select(group=group_name):
            eps.append(ep)
        return eps

elif sys.version_info >= (3,8,0):
    from importlib import metadata

    def _get_entry_points(group_name: str) -> List['EntryPoint']:
        eps = []
        for dist in metadata.distributions():
            for ep in dist.entry_points:
                if ep.group == group_name:
                    eps.append(ep)
        return eps

else:
    import pkg_resources # type: ignore

    def _get_entry_points(group_name: str) -> List['EntryPoint']:
        eps = []
        for ep in pkg_resources.iter_entry_points(group_name):
            eps.append(ep)
        return eps



def get_entry_points(group_name: str) -> List['EntryPoint']:
    return _get_entry_points(group_name)



from typing import TYPE_CHECKING
import re

if TYPE_CHECKING:
    from typing import List, Union, Tuple
    URLRecipe = List[Union[str, Tuple[str, str]]]

class TranslatorSpec:
    """
    Remote URL Patterns
    -------------------
    The ``remote_regexes`` class variable is used to define a list of regular
    expressions to match the remote URL for this service.
    This is typically two regexes - one for SSH and another for HTTPS

    If an expression uses named groups, the group names can be referenced later
    in URL recipes. Since URL patterns in this list are searched sequentially,
    named groups from only the first successful match are captured.


    URL Recipes
    -----------
    URL Recipes are used to construct the resulting weblink. A recipe entry can
    either be simply a format string, or a tuple containing a format string
    followed by a conditional expression:

        recipes = [
            <format string>,
            (<format string>, <expression string>),
        ]

    During recipe processing, keys from an "info" dictionary are made avaialble.
    These include the following:

    * ``path`` - target path, relative to the repository root.
    * ``is_folder`` - Boolean indicating whether the path is a directory or not.
    * ``branch_name`` - The target's current branch name. If not known, this key is not available.
    * ``commit_hash`` - The target's exact commit hash.
    * ``line`` - target line number. Only exists if a single line was specified in the target
    * ``start_line``, ``end_line`` - target line range. Only exists if a range
      of lines were specified in the target.

    .. as well as any named capture groups from the remote URL match earlier.

    Format strings are processed using ``str.format()`` using the keys available
    in the info dictionary.

    Conditional strings are evaluated. Keys from the info dictionary are
    available as local variables.

    The first recipe that matches its conditional (if any), and can be
    formatted successfully (all keys referenced are available) is used.

    """

    #: List of regexes to match the git remote URL.
    remote_regexes = [] # type: List[str]

    #: Formatting recipes for the URL prefix.
    #: This is usually for handling variations in the repository's base URL
    url_root_recipes = [] # type: URLRecipe

    #: Formatting recipes for the main contents of the URL
    #: This is a good place to define variations in file/folder paths,
    #: branch/commit references, etc.
    url_body_recipes = [] # type: URLRecipe

    #: Formatting recipes for the URL's suffix
    #: Use this to define line number highligting, etc.
    url_suffix_recipes = [] # type: URLRecipe

    @classmethod
    def is_match(cls, remote: str) -> bool:
        """
        Checks if remote URL is a match for this translator's service.
        Returning True guarantees that this translator can generate a link.

        Parameters
        ----------
        remote: str
            remote git URL.
        """
        for regex in cls.remote_regexes:
            if re.fullmatch(regex, remote):
                return True

        return False


    @classmethod
    def get_source_url(cls, remote: str, info: dict) -> str:

        # Get match object for remote url
        m = None
        for regex in cls.remote_regexes:
            m = re.fullmatch(regex, remote)
            if m:
                break
        assert m is not None

        # Extract any named match groups and stuff them to the info dictionary
        for k,v in m.groupdict().items():
            if v is not None:
                info[k] = v

        url_root = cls.process_recipe(info, cls.url_root_recipes)
        url_body = cls.process_recipe(info, cls.url_body_recipes)
        url_suffix = cls.process_recipe(info, cls.url_suffix_recipes)

        return url_root + url_body + url_suffix


    @classmethod
    def process_recipe(cls, info: dict, recipes: list) -> str:

        for recipe in recipes:
            if isinstance(recipe, tuple):
                format_string, conditional = recipe

                # Check conditional
                if not eval(conditional, {}, info):
                    continue
            else:
                # unconditional format string
                format_string = recipe
            assert isinstance(format_string, str)

            # Process format string
            try:
                return format_string.format(**info)
            except KeyError:
                # format string tried to use an info key that was unavailable
                # not a match
                pass

        raise RuntimeError



from .translator import TranslatorSpec

class GitLab(TranslatorSpec):
    remote_regexes = [
        r'git@gitlab\.com:(?P<project_name>[\w\-]+)/(?P<repo_name>[\w\-]+)\.git',
        r'https://gitlab\.com/(?P<project_name>[\w\-]+)/(?P<repo_name>[\w\-]+)\.git',
    ]

    url_root_recipes = [
        "https://gitlab.com/{project_name}/{repo_name}/",
    ]

    url_body_recipes = [
        ("blob/{branch_name}/{path}", "is_folder == False"),
        ("tree/{branch_name}/{path}", "is_folder == True"),
        ("blob/{commit_hash}/{path}", "is_folder == False"),
        ("tree/{commit_hash}/{path}", "is_folder == True"),
    ]

    url_suffix_recipes = [
        ("#L{start_line}-L{end_line}", "is_folder == False"),
        ("#L{line}", "is_folder == False"),
        "",
    ]



from typing import TYPE_CHECKING

from .entry_points import get_entry_points

if TYPE_CHECKING:
    from typing import List, Type
    from .translators.translator import TranslatorSpec

def get_translator_plugins() -> 'List[Type[TranslatorSpec]]':
    return [ep.load() for ep in get_entry_points("gitmetheurl.translators")]
